@@ -30,10 +30,10 @@
 #include <usual/strpool.h>
 
 /* regex elements */
-#define WS0	"[ \t\n\r]*"
-#define WS1	"[ \t\n\r]+"
-#define WORD	"(\"([^\"]+|\"\")*\"|[0-9a-z_]+)"
-#define STRING	"('([^']|'')*')"
+#define WS0     "[ \t\n\r]*"
+#define WS1     "[ \t\n\r]+"
+#define WORD    "(\"([^\"]+|\"\")*\"|[0-9a-z_]+)"
+#define STRING  "('([^']|'')*')"
 
 /* possible max + 1 */
 #define MAX_GROUPS 10
@@ -52,15 +52,15 @@ struct cmd_lookup {
 
 /* CMD [arg]; */
 static const char cmd_normal_rx[] =
-"^" WS0 WORD "(" WS1 WORD ")?" WS0 "(;" WS0 ")?$";
+	"^" WS0 WORD "(" WS1 WORD ")?" WS0 "(;" WS0 ")?$";
 
 /* SET with simple value */
 static const char cmd_set_word_rx[] =
-"^" WS0 "set" WS1 WORD WS0 "(=|to)" WS0 WORD WS0 "(;" WS0 ")?$";
+	"^" WS0 "set" WS1 WORD WS0 "(=|to)" WS0 WORD WS0 "(;" WS0 ")?$";
 
 /* SET with quoted value */
 static const char cmd_set_str_rx[] =
-"^" WS0 "set" WS1 WORD WS0 "(=|to)" WS0 STRING WS0 "(;" WS0 ")?$";
+	"^" WS0 "set" WS1 WORD WS0 "(=|to)" WS0 STRING WS0 "(;" WS0 ")?$";
 
 /* compiled regexes */
 static regex_t rc_cmd;
@@ -108,7 +108,7 @@ bool admin_error(PgSocket *admin, const char *fmt, ...)
 
 	log_error("%s", str);
 	if (admin)
-		res = send_pooler_error(admin, true, false, str);
+		res = send_pooler_error(admin, true, NULL, false, str);
 	return res;
 }
 
@@ -159,7 +159,7 @@ bool admin_ready(PgSocket *admin, const char *desc)
 
 /*
  * some silly clients start actively messing with server parameters
- * without checking if thats necessary.  Fake some env for them.
+ * without checking if that's necessary.  Fake some env for them.
  */
 struct FakeParam {
 	const char *name;
@@ -195,8 +195,9 @@ static bool fake_show(PgSocket *admin, const char *name)
 			pktbuf_write_RowDescription(buf, "s", p->name);
 			pktbuf_write_DataRow(buf, "s", p->value);
 			admin_flush(admin, buf, "SHOW");
-		} else
+		} else {
 			admin_error(admin, "no mem");
+		}
 	}
 	return got;
 }
@@ -219,8 +220,9 @@ static bool fake_set(PgSocket *admin, const char *key, const char *val)
 		if (buf) {
 			pktbuf_write_Notice(buf, "SET ignored");
 			admin_flush(admin, buf, "SET");
-		} else
+		} else {
 			admin_error(admin, "no mem");
+		}
 	}
 	return got;
 }
@@ -238,6 +240,9 @@ static bool admin_set(PgSocket *admin, const char *key, const char *val)
 		ok = set_config_param(key, val);
 		if (ok) {
 			PktBuf *buf = pktbuf_dynamic(256);
+			if (!buf) {
+				return admin_error(admin, "no mem");
+			}
 			if (strstr(key, "_tls_") != NULL) {
 				if (!sbuf_tls_setup())
 					pktbuf_write_Notice(buf, "TLS settings could not be applied, still using old configuration");
@@ -247,8 +252,9 @@ static bool admin_set(PgSocket *admin, const char *key, const char *val)
 		} else {
 			return admin_error(admin, "SET failed");
 		}
-	} else
+	} else {
 		return admin_error(admin, "admin access needed");
+	}
 }
 
 /* send a row with sendmsg, optionally attaching a fd */
@@ -276,11 +282,11 @@ static bool send_one_fd(PgSocket *admin,
 	struct PktBuf *pkt = pktbuf_temp();
 
 	pktbuf_write_DataRow(pkt, "issssiqisssssbb",
-		      fd, task, user, db, addr, port, ckey, link,
-		      client_enc, std_strings, datestyle, timezone,
-		      password,
-		      scram_client_key_len, scram_client_key,
-		      scram_server_key_len, scram_server_key);
+			     fd, task, user, db, addr, port, ckey, link,
+			     client_enc, std_strings, datestyle, timezone,
+			     password,
+			     scram_client_key_len, scram_client_key,
+			     scram_server_key_len, scram_server_key);
 	if (pkt->failed)
 		return false;
 	iovec.iov_base = pkt->buf;
@@ -345,19 +351,19 @@ static bool show_one_fd(PgSocket *admin, PgSocket *sk)
 	if (!mbuf_get_uint64be(&tmp, &ckey))
 		return false;
 
-	if (sk->pool && sk->pool->db->auth_user && sk->login_user && !find_user(sk->login_user->name))
-		password = sk->login_user->passwd;
+	if (sk->pool && sk->pool->db->auth_user_credentials && sk->login_user_credentials && !find_global_user(sk->login_user_credentials->name))
+		password = sk->login_user_credentials->passwd;
 
 	/* PAM requires passwords as well since they are not stored externally */
-	if (cf_auth_type == AUTH_PAM && !find_user(sk->login_user->name))
-		password = sk->login_user->passwd;
+	if (cf_auth_type == AUTH_PAM && !find_global_user(sk->login_user_credentials->name))
+		password = sk->login_user_credentials->passwd;
 
-	if (sk->pool && sk->pool->user && sk->pool->user->has_scram_keys)
+	if (sk->pool && sk->pool->user_credentials && sk->pool->user_credentials->has_scram_keys)
 		send_scram_keys = true;
 
 	return send_one_fd(admin, sbuf_socket(&sk->sbuf),
 			   is_server_socket(sk) ? "server" : "client",
-			   sk->login_user ? sk->login_user->name : NULL,
+			   sk->login_user_credentials ? sk->login_user_credentials->name : NULL,
 			   sk->pool ? sk->pool->db->name : NULL,
 			   pga_ntop(addr, addrbuf, sizeof(addrbuf)),
 			   pga_port(addr),
@@ -368,10 +374,10 @@ static bool show_one_fd(PgSocket *admin, PgSocket *sk)
 			   datestyle ? datestyle->str : NULL,
 			   timezone ? timezone->str : NULL,
 			   password,
-			   send_scram_keys ? sk->pool->user->scram_ClientKey : NULL,
-			   send_scram_keys ? (int) sizeof(sk->pool->user->scram_ClientKey) : -1,
-			   send_scram_keys ? sk->pool->user->scram_ServerKey : NULL,
-			   send_scram_keys ? (int) sizeof(sk->pool->user->scram_ServerKey) : -1);
+			   send_scram_keys ? sk->pool->user_credentials->scram_ClientKey : NULL,
+			   send_scram_keys ? (int) sizeof(sk->pool->user_credentials->scram_ClientKey) : -1,
+			   send_scram_keys ? sk->pool->user_credentials->scram_ServerKey : NULL,
+			   send_scram_keys ? (int) sizeof(sk->pool->user_credentials->scram_ServerKey) : -1);
 }
 
 static bool show_pooler_cb(void *arg, int fd, const PgAddr *a)
@@ -404,19 +410,6 @@ static bool show_fds_from_list(PgSocket *admin, struct StatList *list)
 	return res;
 }
 
-static PgDatabase *find_or_register_database(PgSocket *admin, const char *name)
-{
-	PgDatabase *db = find_database(name);
-	if (db == NULL) {
-		db = register_auto_database(name);
-		if (db != NULL) {
-			slog_info(admin,
-				  "registered new auto-database: %s", name);
-		}
-	}
-	return db;
-}
-
 /*
  * Command: SHOW FDS
  *
@@ -447,13 +440,13 @@ static bool admin_show_fds(PgSocket *admin, const char *arg)
 	 * send resultset
 	 */
 	SEND_RowDescription(res, admin, "issssiqisssssbb",
-				 "fd", "task",
-				 "user", "database",
-				 "addr", "port",
-				 "cancel", "link",
-				 "client_encoding", "std_strings",
-				 "datestyle", "timezone", "password",
-				 "scram_client_key", "scram_server_key");
+			    "fd", "task",
+			    "user", "database",
+			    "addr", "port",
+			    "cancel", "link",
+			    "client_encoding", "std_strings",
+			    "datestyle", "timezone", "password",
+			    "scram_client_key", "scram_server_key");
 	if (res)
 		res = show_pooler_fds(admin);
 
@@ -492,6 +485,7 @@ static bool admin_show_databases(PgSocket *admin, const char *arg)
 	PktBuf *buf;
 	struct CfValue cv;
 	const char *pool_mode_str;
+	usec_t server_lifetime_secs;
 
 	cv.extra = pool_mode_map;
 	buf = pktbuf_dynamic(256);
@@ -500,24 +494,29 @@ static bool admin_show_databases(PgSocket *admin, const char *arg)
 		return true;
 	}
 
-	pktbuf_write_RowDescription(buf, "ssissiiisiiii",
+	pktbuf_write_RowDescription(buf, "ssissiiiisiiii",
 				    "name", "host", "port",
 				    "database", "force_user", "pool_size", "min_pool_size", "reserve_pool",
-				    "pool_mode", "max_connections", "current_connections", "paused", "disabled");
+				    "server_lifetime", "pool_mode", "max_connections", "current_connections",
+				    "paused", "disabled");
 	statlist_for_each(item, &database_list) {
 		db = container_of(item, PgDatabase, head);
 
-		f_user = db->forced_user ? db->forced_user->name : NULL;
+		server_lifetime_secs = (db->server_lifetime > 0 ? db->server_lifetime : cf_server_lifetime) / USEC;
+		f_user = db->forced_user_credentials ? db->forced_user_credentials->name : NULL;
 		pool_mode_str = NULL;
 		cv.value_p = &db->pool_mode;
 		if (db->pool_mode != POOL_INHERIT)
 			pool_mode_str = cf_get_lookup(&cv);
-		pktbuf_write_DataRow(buf, "ssissiiisiiii",
+
+
+		pktbuf_write_DataRow(buf, "ssissiiiisiiii",
 				     db->name, db->host, db->port,
 				     db->dbname, f_user,
 				     db->pool_size >= 0 ? db->pool_size : cf_default_pool_size,
 				     db->min_pool_size >= 0 ? db->min_pool_size : cf_min_pool_size,
 				     db->res_pool_size >= 0 ? db->res_pool_size : cf_res_pool_size,
+				     server_lifetime_secs,
 				     pool_mode_str,
 				     database_max_connections(db),
 				     db->connection_count,
@@ -590,10 +589,10 @@ static bool admin_show_lists(PgSocket *admin, const char *arg)
 /* Command: SHOW USERS */
 static bool admin_show_users(PgSocket *admin, const char *arg)
 {
-	PgUser *user;
 	struct List *item;
 	PktBuf *buf = pktbuf_dynamic(256);
 	struct CfValue cv;
+	char pool_size_str[12] = "";
 	const char *pool_mode_str;
 
 	if (!buf) {
@@ -602,33 +601,42 @@ static bool admin_show_users(PgSocket *admin, const char *arg)
 	}
 	cv.extra = pool_mode_map;
 
-	pktbuf_write_RowDescription(buf, "ss", "name", "pool_mode");
+	pktbuf_write_RowDescription(buf, "sssii", "name", "pool_size", "pool_mode", "max_user_connections", "current_connections");
 	statlist_for_each(item, &user_list) {
-		user = container_of(item, PgUser, head);
+		PgGlobalUser *user = container_of(item, PgGlobalUser, head);
+		if (user->pool_size >= 0)
+			snprintf(pool_size_str, sizeof(pool_size_str), "%9d", user->pool_size);
 		pool_mode_str = NULL;
+
 		cv.value_p = &user->pool_mode;
 		if (user->pool_mode != POOL_INHERIT)
 			pool_mode_str = cf_get_lookup(&cv);
 
-		pktbuf_write_DataRow(buf, "ss", user->name, pool_mode_str);
+		pktbuf_write_DataRow(buf, "sssii", user->credentials.name,
+				     pool_size_str,
+				     pool_mode_str,
+				     user_max_connections(user),
+				     user->connection_count
+				     );
 	}
 	admin_flush(admin, buf, "SHOW");
 	return true;
 }
 
-#define SKF_STD "sssssisiTTiiississ"
-#define SKF_DBG "sssssisiTTiiississiiiiiii"
+#define SKF_STD "ssssssisiTTiiississi"
+#define SKF_DBG "ssssssisiTTiiississiiiiiiii"
 
 static void socket_header(PktBuf *buf, bool debug)
 {
 	pktbuf_write_RowDescription(buf, debug ? SKF_DBG : SKF_STD,
-				    "type", "user", "database", "state",
+				    "type", "user", "database", "replication", "state",
 				    "addr", "port", "local_addr", "local_port",
 				    "connect_time", "request_time",
 				    "wait", "wait_us", "close_needed",
 				    "ptr", "link", "remote_pid", "tls",
 				    "application_name",
-				    /* debug follows */
+				    "prepared_statements",
+					/* debug follows */
 				    "recv_pos", "pkt_pos", "pkt_remain",
 				    "send_pos", "send_remain",
 				    "pkt_avail", "send_avail");
@@ -643,6 +651,7 @@ static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
 {
 	int pkt_avail = 0, send_avail = 0;
 	int remote_pid;
+	int prepared_statement_count = 0;
 	char ptrbuf[128], linkbuf[128];
 	char l_addr[PGADDR_BUF], r_addr[PGADDR_BUF];
 	IOBuf *io = sk->sbuf.io;
@@ -651,6 +660,7 @@ static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
 	const struct PStr *application_name = v->var_list[VAppName];
 	usec_t now = get_cached_time();
 	usec_t wait_time = sk->query_start ? now - sk->query_start : 0;
+	char *replication;
 
 	if (io) {
 		pkt_avail = iobuf_amount_parse(sk->sbuf.io);
@@ -678,10 +688,23 @@ static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
 	if (sk->sbuf.tls)
 		tls_get_connection_info(sk->sbuf.tls, infobuf, sizeof infobuf);
 
+	if (is_server_socket(sk))
+		prepared_statement_count = HASH_COUNT(sk->server_prepared_statements);
+	else
+		prepared_statement_count = HASH_COUNT(sk->client_prepared_statements);
+
+	if (sk->replication == REPLICATION_NONE)
+		replication = "none";
+	else if (sk->replication == REPLICATION_LOGICAL)
+		replication = "logical";
+	else
+		replication = "physical";
+
 	pktbuf_write_DataRow(buf, debug ? SKF_DBG : SKF_STD,
-			     is_server_socket(sk) ? "S" :"C",
-			     sk->login_user ? sk->login_user->name : "(nouser)",
+			     is_server_socket(sk) ? "S" : "C",
+			     sk->login_user_credentials ? sk->login_user_credentials->name : "(nouser)",
 			     sk->pool && !sk->pool->db->peer_id ? sk->pool->db->name : "(nodb)",
+			     replication,
 			     state, r_addr, pga_port(&sk->remote_addr),
 			     l_addr, pga_port(&sk->local_addr),
 			     sk->connect_time,
@@ -691,7 +714,8 @@ static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
 			     sk->close_needed,
 			     ptrbuf, linkbuf, remote_pid, infobuf,
 			     application_name ? application_name->str : "",
-			     /* debug */
+			     prepared_statement_count,
+				/* debug */
 			     io ? io->recv_pos : 0,
 			     io ? io->parse_pos : 0,
 			     sk->sbuf.pkt_remain,
@@ -883,9 +907,9 @@ static bool admin_show_pools(PgSocket *admin, const char *arg)
 		pool = container_of(item, PgPool, head);
 		waiter = first_socket(&pool->waiting_client_list);
 		max_wait = (waiter && waiter->query_start) ? now - waiter->query_start : 0;
-		pool_mode = pool_pool_mode(pool);
+		pool_mode = probably_wrong_pool_pool_mode(pool);
 		pktbuf_write_DataRow(buf, "ssiiiiiiiiiiiiis",
-				     pool->db->name, pool->user->name,
+				     pool->db->name, pool->user_credentials->name,
 				     statlist_count(&pool->active_client_list),
 				     statlist_count(&pool->waiting_client_list),
 				     statlist_count(&pool->active_cancel_req_list),
@@ -897,7 +921,7 @@ static bool admin_show_pools(PgSocket *admin, const char *arg)
 				     statlist_count(&pool->used_server_list),
 				     statlist_count(&pool->tested_server_list),
 				     statlist_count(&pool->new_server_list),
-				     /* how long is the oldest client waited */
+					/* how long is the oldest client waited */
 				     (int)(max_wait / USEC),
 				     (int)(max_wait % USEC),
 				     cf_get_lookup(&cv));
@@ -1099,8 +1123,15 @@ static bool admin_cmd_reload(PgSocket *admin, const char *arg)
 /* Command: SHUTDOWN */
 static bool admin_cmd_shutdown(PgSocket *admin, const char *arg)
 {
-	if (arg && *arg)
-		return syntax_error(admin);
+	enum ShutDownMode mode = SHUTDOWN_IMMEDIATE;
+	if (arg && *arg) {
+		if (strcasecmp(arg, "WAIT_FOR_CLIENTS") == 0)
+			mode = SHUTDOWN_WAIT_FOR_CLIENTS;
+		else if (strcasecmp(arg, "WAIT_FOR_SERVERS") == 0)
+			mode = SHUTDOWN_WAIT_FOR_SERVERS;
+		else
+			return syntax_error(admin);
+	}
 
 	if (!admin->admin_user)
 		return admin_error(admin, "admin access needed");
@@ -1110,11 +1141,25 @@ static bool admin_cmd_shutdown(PgSocket *admin, const char *arg)
 	 * event from fd.  Currently atexit() cleanup should be called
 	 * before closing open sockets.
 	 */
-	log_info("SHUTDOWN command issued");
-	cf_shutdown = 2;
-	event_base_loopbreak(pgb_event_base);
-
-	return true;
+	cf_shutdown = mode;
+	if (mode == SHUTDOWN_IMMEDIATE) {
+		log_info("SHUTDOWN command issued");
+		event_base_loopbreak(pgb_event_base);
+		/*
+		 * By not running admin_ready the connection is kept open
+		 * until the process is actually shut down.
+		 */
+		return true;
+	} else {
+		if (mode == SHUTDOWN_WAIT_FOR_SERVERS) {
+			cf_pause_mode = P_PAUSE;
+			log_info("SHUTDOWN WAIT_FOR_SERVERS command issued");
+		} else {
+			log_info("SHUTDOWN WAIT_FOR_CLIENTS command issued");
+		}
+		cleanup_sockets();
+		return admin_ready(admin, "SHUTDOWN");
+	}
 }
 
 static void full_resume(void)
@@ -1123,12 +1168,6 @@ static void full_resume(void)
 	cf_pause_mode = P_NONE;
 	if (tmp_mode == P_SUSPEND)
 		resume_all();
-
-	/* avoid surprise later if cf_shutdown stays set */
-	if (cf_shutdown) {
-		log_info("canceling shutdown");
-		cf_shutdown = 0;
-	}
 }
 
 /* Command: RESUME */
@@ -1139,10 +1178,13 @@ static bool admin_cmd_resume(PgSocket *admin, const char *arg)
 
 	if (!arg[0]) {
 		log_info("RESUME command issued");
-		if (cf_pause_mode != P_NONE)
+		if (cf_shutdown) {
+			return admin_error(admin, "pooler is shutting down");
+		} else if (cf_pause_mode != P_NONE) {
 			full_resume();
-		else
+		} else {
 			return admin_error(admin, "pooler is not paused/suspended");
+		}
 	} else {
 		PgDatabase *db = find_database(arg);
 		log_info("RESUME '%s' command issued", arg);
@@ -1327,41 +1369,41 @@ static bool admin_cmd_wait_close(PgSocket *admin, const char *arg)
 	if (!admin->admin_user)
 		return admin_error(admin, "admin access needed");
 
-       if (!arg[0]) {
-	       struct List *item;
-	       PgPool *pool;
-	       int active = 0;
+	if (!arg[0]) {
+		struct List *item;
+		PgPool *pool;
+		int active = 0;
 
-	       log_info("WAIT_CLOSE command issued");
-	       statlist_for_each(item, &pool_list) {
-		       PgDatabase *db;
+		log_info("WAIT_CLOSE command issued");
+		statlist_for_each(item, &pool_list) {
+			PgDatabase *db;
 
-		       pool = container_of(item, PgPool, head);
-		       db = pool->db;
-		       db->db_wait_close = true;
-		       active += count_db_active(db);
-	       }
-	       if (active > 0)
-		       admin->wait_for_response = true;
-	       else
-		       return admin_ready(admin, "WAIT_CLOSE");
-       } else {
-	       PgDatabase *db;
+			pool = container_of(item, PgPool, head);
+			db = pool->db;
+			db->db_wait_close = true;
+			active += count_db_active(db);
+		}
+		if (active > 0)
+			admin->wait_for_response = true;
+		else
+			return admin_ready(admin, "WAIT_CLOSE");
+	} else {
+		PgDatabase *db;
 
-	       log_info("WAIT_CLOSE '%s' command issued", arg);
-	       db = find_or_register_database(admin, arg);
-	       if (db == NULL)
-		       return admin_error(admin, "no such database: %s", arg);
-	       if (db == admin->pool->db)
-		       return admin_error(admin, "cannot wait in admin db: %s", arg);
-	       db->db_wait_close = true;
-	       if (count_db_active(db) > 0)
-		       admin->wait_for_response = true;
-	       else
-		       return admin_ready(admin, "WAIT_CLOSE");
-       }
+		log_info("WAIT_CLOSE '%s' command issued", arg);
+		db = find_or_register_database(admin, arg);
+		if (db == NULL)
+			return admin_error(admin, "no such database: %s", arg);
+		if (db == admin->pool->db)
+			return admin_error(admin, "cannot wait in admin db: %s", arg);
+		db->db_wait_close = true;
+		if (count_db_active(db) > 0)
+			admin->wait_for_response = true;
+		else
+			return admin_ready(admin, "WAIT_CLOSE");
+	}
 
-       return true;
+	return true;
 }
 
 /* extract substring from regex group */
@@ -1393,7 +1435,7 @@ static bool copy_arg(const char *src, regmatch_t *glist,
 	/* copy and unquote */
 	if (*s == qchar) {
 		for (i = 1; i < len - 1; i++) {
-			if (s[i] == qchar && s[i+1] == qchar)
+			if (s[i] == qchar && s[i + 1] == qchar)
 				i++;
 			*d++ = s[i];
 		}
@@ -1409,25 +1451,26 @@ static bool admin_show_help(PgSocket *admin, const char *arg)
 {
 	bool res;
 	SEND_generic(res, admin, 'N',
-		"sssss",
-		"SNOTICE", "C00000", "MConsole usage",
-		"D\n\tSHOW HELP|CONFIG|DATABASES"
-		"|POOLS|CLIENTS|SERVERS|USERS|VERSION\n"
-		"\tSHOW PEERS|PEER_POOLS\n"
-		"\tSHOW FDS|SOCKETS|ACTIVE_SOCKETS|LISTS|MEM|STATE\n"
-		"\tSHOW DNS_HOSTS|DNS_ZONES\n"
-		"\tSHOW STATS|STATS_TOTALS|STATS_AVERAGES|TOTALS\n"
-		"\tSET key = arg\n"
-		"\tRELOAD\n"
-		"\tPAUSE [<db>]\n"
-		"\tRESUME [<db>]\n"
-		"\tDISABLE <db>\n"
-		"\tENABLE <db>\n"
-		"\tRECONNECT [<db>]\n"
-		"\tKILL <db>\n"
-		"\tSUSPEND\n"
-		"\tSHUTDOWN\n"
-		"\tWAIT_CLOSE [<db>]", "");
+		     "sssss",
+		     "SNOTICE", "C00000", "MConsole usage",
+		     "D\n\tSHOW HELP|CONFIG|DATABASES"
+		     "|POOLS|CLIENTS|SERVERS|USERS|VERSION\n"
+		     "\tSHOW PEERS|PEER_POOLS\n"
+		     "\tSHOW FDS|SOCKETS|ACTIVE_SOCKETS|LISTS|MEM|STATE\n"
+		     "\tSHOW DNS_HOSTS|DNS_ZONES\n"
+		     "\tSHOW STATS|STATS_TOTALS|STATS_AVERAGES|TOTALS\n"
+		     "\tSET key = arg\n"
+		     "\tRELOAD\n"
+		     "\tPAUSE [<db>]\n"
+		     "\tRESUME [<db>]\n"
+		     "\tDISABLE <db>\n"
+		     "\tENABLE <db>\n"
+		     "\tRECONNECT [<db>]\n"
+		     "\tKILL <db>\n"
+		     "\tSUSPEND\n"
+		     "\tSHUTDOWN\n"
+		     "\tSHUTDOWN WAIT_FOR_SERVERS|WAIT_FOR_CLIENTS\n"
+		     "\tWAIT_CLOSE [<db>]", "");
 	if (res)
 		res = admin_ready(admin, "SHOW");
 	return res;
@@ -1557,8 +1600,9 @@ static bool admin_parse_query(PgSocket *admin, const char *q)
 		if (!ok)
 			goto failed;
 		res = admin_set(admin, arg, val);
-	} else
+	} else {
 		res = syntax_error(admin);
+	}
 done:
 	current_query = NULL;
 	if (!res)
@@ -1630,12 +1674,12 @@ bool admin_pre_login(PgSocket *client, const char *username)
 
 		res = getpeereid(sbuf_socket(&client->sbuf), &peer_uid, &peer_gid);
 		if (res >= 0 && peer_uid == getuid()
-			&& strcmp("pgbouncer", username) == 0)
-		{
-			client->login_user = admin_pool->db->forced_user;
+		    && strcmp("pgbouncer", username) == 0) {
+			client->login_user_credentials = admin_pool->db->forced_user_credentials;
 			client->own_user = true;
 			client->admin_user = true;
-			slog_info(client, "pgbouncer access from unix socket");
+			if (cf_log_connections)
+				slog_info(client, "pgbouncer access from unix socket");
 			return true;
 		}
 	}
@@ -1646,11 +1690,11 @@ bool admin_pre_login(PgSocket *client, const char *username)
 	 */
 	if (cf_auth_type == AUTH_ANY) {
 		if (strlist_contains(cf_admin_users, username)) {
-			client->login_user = admin_pool->db->forced_user;
+			client->login_user_credentials = admin_pool->db->forced_user_credentials;
 			client->admin_user = true;
 			return true;
 		} else if (strlist_contains(cf_stats_users, username)) {
-			client->login_user = admin_pool->db->forced_user;
+			client->login_user_credentials = admin_pool->db->forced_user_credentials;
 			return true;
 		}
 	}
@@ -1659,7 +1703,7 @@ bool admin_pre_login(PgSocket *client, const char *username)
 
 bool admin_post_login(PgSocket *client)
 {
-	const char *username = client->login_user->name;
+	const char *username = client->login_user_credentials->name;
 
 	if (cf_auth_type == AUTH_ANY)
 		return true;
@@ -1680,7 +1724,7 @@ void admin_setup(void)
 {
 	PgDatabase *db;
 	PgPool *pool;
-	PgUser *user;
+	PgGlobalUser *user;
 	PktBuf *msg;
 	int res;
 
@@ -1693,20 +1737,20 @@ void admin_setup(void)
 	db->pool_size = 2;
 	db->admin = true;
 	db->pool_mode = POOL_STMT;
-	if (!force_user(db, "pgbouncer", ""))
+	if (!force_user_credentials(db, "pgbouncer", ""))
 		die("no mem on startup - cannot alloc pgbouncer user");
 
 	/* fake pool */
-	pool = get_pool(db, db->forced_user);
+	pool = get_pool(db, db->forced_user_credentials);
 	if (!pool)
 		die("cannot create admin pool?");
 	admin_pool = pool;
 
 	/* user */
-	user = find_user("pgbouncer");
+	user = find_global_user("pgbouncer");
 	if (!user) {
 		/* fake user with disabled psw */
-		user = add_user("pgbouncer", "");
+		user = add_global_user("pgbouncer", "");
 		if (!user)
 			die("cannot create admin user?");
 	}
@@ -1769,9 +1813,9 @@ void admin_pause_done(void)
 			res = admin_ready(admin, "SUSPEND");
 			break;
 		default:
-			if (count_paused_databases() > 0)
+			if (count_paused_databases() > 0) {
 				res = admin_ready(admin, "PAUSE");
-			else {
+			} else {
 				/* FIXME */
 				fatal("admin_pause_done: bad state");
 				res = false;
@@ -1785,8 +1829,7 @@ void admin_pause_done(void)
 	}
 
 	if (statlist_empty(&admin_pool->active_client_list)
-	    && cf_pause_mode == P_SUSPEND)
-	{
+	    && cf_pause_mode == P_SUSPEND) {
 		log_info("admin disappeared when suspended, doing RESUME");
 		cf_pause_mode = P_NONE;
 		resume_all();
@@ -1820,6 +1863,9 @@ void admin_handle_cancel(PgSocket *admin)
 	if (!admin->wait_for_response)
 		slog_warning(admin, "admin cancel request for non-waiting client?");
 
+	if (cf_shutdown)
+		return;
+
 	if (cf_pause_mode != P_NONE)
 		full_resume();
 }
